"""spatialhash.py - High performance spatial hash for spatial
partitioning and fast collision detection.

Objects (other than geometry.LineGeometry) must have a pygame Rect attribute.
Optionally, objects may have a collided static method attribute for lower-level
collision detection (see the gummworld2.geometry module).

Objects that are outside the world bounding rect are ignored by add().

As of v0.5.0 SpatialHash will sort objects returned by most methods that return
a list of objects. See the methods' docstrings. In order to have the objects
sorted set the SpatialHash's sort_key method. This is the same key that is
passed to sort(key=func) and sorted(key=func), e.g.:
key=lambda obj: obj.rect.bottom.

This module is derived from the article and source code written by Conkerjo at
http://conkerjo.wordpress.com/2009/06/13/spatial-hashing-implementation-for-fast-2d-collisions/.
"""

from math import ceil
# from weakref import WeakKeyDictionary
from collections import defaultdict

import gsdl2
from gsdl2 import Rect

import geometry


__all__ = ['SpatialHash']


class SpatialHash(object):
    
    def __init__(self, world_rect, cell_size):
        ## Must inflate world rect by 1. This is because pygame rects consider
        ## the right and bottom borders to have zero units width. But the
        ## hashing calculates those as one unit width. Thus, taking pygame
        ## rect space at face value will attempt to access non-existant buckets
        ## at the corner cases.
        world_rect = world_rect.inflate(1, 1)
        
        self.rect = Rect(world_rect)
        self.bounds = (
            world_rect[0],
            world_rect[1],
            world_rect[0] + world_rect[2],
            world_rect[1] + world_rect[3],
        )
        self.cell_size = int(cell_size)
        
        self.rows = int(ceil(world_rect.h / float(cell_size))) + 1
        self.cols = int(ceil(world_rect.w / float(cell_size))) + 1
        self.buckets = [[] for i in range(self.rows * self.cols)]
        ## cell_ids = {obj1:[cells...], obj2:[cells...], ...}
        self.cell_ids = {}  # WeakKeyDictionary()

        self.num_buckets = len(self.buckets)
        
        self.coll_tests = 0
        self._temp_rect = gsdl2.Rect(0, 0, 0, 0)
        
        self.sort_key = None
        self.dirty = False
        self.sorted_objects = []
    
    @property
    def objects(self):
        """Return the entire list of objects.
        
        This method honors self.sort_key.
        """
        if self.sort_key:
            if self.dirty:
                self.sorted_objects[:] = sorted(
                    self.cell_ids, key=self.sort_key)
                self.dirty = False
            return self.sorted_objects[:]
        else:
            return self.cell_ids.keys()
    
    def add(self, obj):
        """Add or re-add obj. Return True if in bounds, else return False.
        
        If this method returns False then the object is completely out of
        bounds and cannot be stored in this space.
        
        Note that when obj changes its position, you must add it again so that
        its cell membership is updated. This method first removes the object if
        it is already in the spatial hash.
        """
        self.dirty = True
        try:
            cell_ids = self.intersect_indices(obj.rect)
        except AttributeError:
            rect = self._get_rect_for_line(obj)
            cell_ids = self.intersect_indices(rect)
        if obj in self.cell_ids and cell_ids == self.cell_ids[obj]:
            return cell_ids == True
        self.remove(obj)
        buckets = self.buckets
        for idx in cell_ids:
            buckets[idx].append(obj)
        self.cell_ids[obj] = cell_ids
        return cell_ids == True
    
    def addlist(self, objs):
        self.dirty = True
        all_good = True
        intersect_indices = self.intersect_indices
        _get_rect_for_line = self._get_rect_for_line
        self_cell_ids = self.cell_ids
        remove = self.remove
        buckets = self.buckets
        for obj in objs:
            try:
                cell_ids = intersect_indices(obj.rect)
            except AttributeError:
                rect = _get_rect_for_line(obj)
                cell_ids = intersect_indices(rect)
            if obj in self_cell_ids and cell_ids == self_cell_ids[obj]:
                # obj is still in same cell ids; save some CPU
                continue
            remove(obj)
            for idx in cell_ids:
                try:
                    buckets[idx].append(obj)
                except IndexError:
                    print('IndexError: buckets[{0}]'.format(idx))
                    print('len(buckets): {0}'.format(len(buckets)))
            self_cell_ids[obj] = cell_ids
            all_good = all_good and cell_ids == True
        return all_good
    
    def remove(self, obj):
        """Remove obj.
        """
        self.dirty = True
        buckets = self.buckets
        cell_ids = self.cell_ids
        if obj in cell_ids:
            for cell_id in tuple(cell_ids[obj]):
                buckets[cell_id].remove(obj)
            del cell_ids[obj]
    
    def get_nearby_objects(self, obj):
        """Return a list of objects that share the same cells as obj.
        
        This method honors self.sort_key.
        """
        nearby_objs = []
        try:
            cell_ids = self.intersect_indices(obj.rect)
        except AttributeError:
            rect = self._get_rect_for_line(obj)
            cell_ids = self.intersect_indices(rect)
        buckets = self.buckets
        for cell_id in cell_ids:
            nearby_objs.extend(buckets[cell_id])
        nearby_objs = list(set(nearby_objs))
        if self.sort_key:
            nearby_objs.sort(key=self.sort_key)
        return nearby_objs
    
    def get_cell(self, cell_id):
        """Return the cell stored at bucket index cell_id.
        
        The returned cell is a list of objects. None is returned if a cell does
        not exist for cell_id.
        """
        try:
            return self.buckets[cell_id]
        except:
            return []
    
    def index(self, cell):
        """Return the bucket index of cell.
        
        Returns None if cell does not exist in buckets.
        
        Note that SpatialHash.buckets.index(cell) does *NOT* work because
        list.index() tests equality, not identity.
        """
        for i, c in enumerate(self.buckets):
            if c is cell:
                return i
    
    def index_at(self, x, y):
        """Return the cell_id of the cell that contains point (x,y).
        
        None is returned if point (x,y) is not in bounds.
        """
        cell_size = self.cell_size
        rect = self.rect
        idx = ((x - rect[0]) // cell_size) + ((y - rect[1]) // cell_size) * self.cols
        return int(idx) if -1 < idx < self.num_buckets else None
    
    def intersect_indices(self, rect):
        """Return list of cell ids that intersect rect.
        """
        # Not pretty, but these ugly optimizations shave 50% off run-time
        # versus function calls and attributes. This method is called by add(),
        # which gets called whenever an object moves.
        
        # return value
        cell_ids = {}
        
        # pre-calculate bounds
        left = rect[0]
        top = rect[1]
        right = left + rect[2]
        bottom = top + rect[3]
        wl, wt, wr, wb = self.bounds
        if left < wl:
            left = wl
        if top < wt:
            top = wt
        if right > wr:
            right = wr
        if bottom > wb:
            bottom = wb
        cell_size = self.cell_size
        
        # pre-calculate loop ranges
        lrange = range
        x_range = list(lrange(left, right, cell_size)) + [right]
        y_range = list(lrange(top, bottom, cell_size)) + [bottom]
        
        # misc speedups
        cols = self.cols
        
        for x in x_range:
            for y in y_range:
                cell_id = x // cell_size + y // cell_size * cols
                cell_ids[cell_id] = 1
        
#        return list(cell_ids)
        return cell_ids.keys()
    
    def intersect_objects(self, rect):
        """Return list of objects whose rects intersect rect.
        
        This method honors self.sort_key.
        """
        objs = {}
        colliderect = rect.colliderect
        rg = geometry.RectGeometry(*rect)
        for cell_id in self.intersect_indices(rect):
            for o in self.get_cell(cell_id):
                try:
                    if colliderect(o.rect):
                        try:
                            if o.collided(o, rg, True):
                                objs[o] = 1
                        except AttributeError:
                            objs[o] = 1
                except AttributeError:
                    try:
                        if o.collided(o, rg, True):
                            objs[o] = 1
                    except AttributeError:
                        pass
        objs = list(objs)
        if self.sort_key:
            objs.sort(key=self.sort_key)
        return objs
    
    def get_cell_grid(self, cell_id):
        """Return the (col,row) coordinate for cell id.
        """
        cols = self.cols
        x = cell_id // cols
        y = cell_id - x * cols
        return x, y
    
    def get_cell_pos(self, cell_id):
        """Return the world coordinates for topleft corner of cell.
        """
        x, y = self.get_cell_grid(cell_id)
        cell_size = self.cell_size
        rect = self.rect
        return x * cell_size + rect.left, y * cell_size + rect.top
    
    def collideany(self, obj):
        """Return True if obj collides with any other object, else False.
        """
        for other in self.get_nearby_objects(obj):
            if other is obj:
                continue
            if hasattr(obj, 'rect') and hasattr(other, 'rect'):
                if obj.rect.colliderect(other.rect):
                    if self._ext_collided(obj, other, True):
                        return True
            else:
                if self._ext_collided(obj, other, True):
                    return True
        return False
    
    def collide(self, obj):
        """Return list of objects that collide with obj.
        """
        collisions = []
        append = collisions.append
        for other in self.get_nearby_objects(obj):
            if other is obj:
                continue
            if hasattr(obj, 'rect') and hasattr(other, 'rect'):
                if obj.rect.colliderect(other.rect):
                    if self._ext_collided(obj, other, True):
                        append(other)
            else:
                if self._ext_collided(obj, other):
                    append(other)
        return collisions
    
    def collidealldict(self, rect=None):
        """Return dict of all collisions.
        
        If rect is specified, only the cells that intersect rect will be
        checked.
        
        The contents of the returned dict are: {obj : [other1,other2,...],...}
        """
        collisions = defaultdict(set)
        self.coll_tests = 0
        if rect:
            buckets = self.buckets
            cells = [buckets[i] for i in self.intersect_indices(rect)]
        else:
            cells = self.buckets
        tests = 0
        for cell in cells:
            for obj in cell:
                for other in cell:
                    if other is obj:
                        continue
                    tests += 1
                    if hasattr(obj, 'rect') and hasattr(other, 'rect'):
                        if obj.rect.colliderect(other.rect):
                            if self._ext_collided(obj, other, True):
                                collisions[obj].add(other)
                    else:
                        if self._ext_collided(obj, other):
                            collisions[obj].add(other)
        self.coll_tests = tests
        return collisions
    
    def collidealllist(self, rect=None):
        """Return list of all collisions.
        
        If rect is specified, only the cells that intersect rect will be
        checked.
        
        The contents of the returned list are: [(obj,other),...]
        """
        collisions = {}
        if rect:
            buckets = self.buckets
            cells = [buckets[i] for i in self.intersect_indices(rect)]
        else:
            cells = self.buckets
        tests = 0
        for cell in cells:
            for obj in cell:
                for other in cell:
                    if other is obj:
                        continue
                    tests += 1
                    if hasattr(obj, 'rect') and hasattr(other, 'rect'):
                        if obj.rect.colliderect(other.rect):
                            if self._ext_collided(obj, other, True):
                                collisions[(obj, other)] = 1
                    else:
                        if self._ext_collided(obj, other):
                            collisions[(obj, other)] = 1
        self.coll_tests = tests
        return collisions.keys()

    def collideallflatlist(self, rect=None):
        """Return flat list of all collisions.
        
        If rect is specified, only the cells that intersect rect will be
        checked.
        
        The contents of the returned list are: [obj1,other1,obj2,other2...]
        """
        collisions = []
        append = collisions.append
        if rect:
            buckets = self.buckets
            cells = [buckets[i] for i in self.intersect_indices(rect)]
        else:
            cells = self.buckets
        tests = 0
        for cell in cells:
            for obj in cell:
                for other in cell:
                    if other is obj:
                        continue
                    tests += 1
                    if hasattr(obj, 'rect') and hasattr(other, 'rect'):
                        if obj.rect.colliderect(other.rect):
                            if self._ext_collided(obj, other, True):
                                append(obj)
                                append(other)
                    else:
                        if self._ext_collided(obj, other):
                            append(obj)
                            append(other)
        self.coll_tests = tests
        return collisions

    @staticmethod
    def _ext_collided(obj, other, rect_collided=False):
        try:
            return obj.collided(obj, other, rect_collided)
        except AttributeError:
            return rect_collided
        # if hasattr(obj, 'collided'):
        #     if hasattr(other, 'collided') or hasattr(other, 'rect'):
        #         return obj.collided(obj, other, rect_collided)
        # elif hasattr(obj, 'rect'):
        #     if hasattr(other, 'collided') or hasattr(other, 'rect'):
        #         return obj.collided(obj, other, rect_collided)
        # return rect_collided
        # if hasattr(obj, 'collided') and hasattr(other, 'collided'):
        #     return obj.collided(obj, other, rect_collided)
        # elif hasattr(obj, 'collided') and hasattr(other, 'rect'):
        #     return obj.collided(obj, other, rect_collided)
        # elif hasattr(obj, 'rect') and hasattr(other, 'collided'):
        #     return obj.collided(obj, other, rect_collided)
        # return rect_collided
        # if rect_collided:
        #     if hasattr(obj, 'collided'):
        #         if obj.collided(obj, other, rect_collided):
        #             return True
        # else:
        #     if hasattr(obj, 'collided') and hasattr(other, 'collided'):
        #         if other.collided(other, obj, rect_collided):
        #             return True
        # return rect_collided

    def clear(self):
        """Clear all objects.
        """
        self.dirty = True
        for cell in self.buckets:
            del cell[:]
        self.cell_ids.clear()
        del self.sorted_objects[:]
    
    def iterobjects(self):
        """Returns a generator that iterates over all objects.
        
        Invoking a SpatialHash object as an iterator produces the same behavior
        as iterobjects().
        
        This method honors self.sort_key.
        """
        if self.sort_key:
            if self.dirty:
                self.sorted_objects[:] = sorted(
                    self.cell_ids, key=self.sort_key)
                self.dirty = False
            for obj in self.sorted_objects:
                yield obj
        else:
            for obj in self.cell_ids:
                yield obj
    
    def itercells(self):
        """Returns a generator that iterates over all cells.
        """
        for cell in self.buckets:
            yield cell
    
    def _get_rect_for_line(self, obj):
        """Lines don't have a rect attribute. Use self._temp_rect to fudge it.
        """
        points = obj.points
        rect = self._temp_rect
        x1, y1 = points[0]
        x2, y2 = points[1]
        if x1 > x2:
            t = x1
            x1 = x2
            x2 = t
        if y1 > y2:
            t = y1
            y1 = y2
            y2 = t
        rect.topleft = x1, y1
        rect.width = x2 - x1
        rect.height = y2 - y1
        return rect
    
    def __iter__(self):
        """This method honors self.sort_key."""
        if self.sort_key:
            if self.dirty:
                self.sorted_objects[:] = sorted(self.cell_ids, key=self.sort_key)
                self.dirty = False
            for obj in self.sorted_objects:
                yield obj
        else:
            for obj in self.cell_ids:
                yield obj
    
    def __contains__(self, obj):
        return obj in self.cell_ids
    
    def __len__(self):
        return len(self.cell_ids)
    
    def __str__(self):
        return '<{}({},{})>'.format(self.__class__.__name__, str(self.rect), str(self.cell_size))

    def __repr(self):
        return self.__str__()
